#include <assert.h>
#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include "ppapi/c/pp_bool.h"
#include "ppapi/c/pp_completion_callback.h"
#include "ppapi/c/pp_errors.h"
#include "ppapi/c/pp_graphics_3d.h"
#include "ppapi/c/pp_input_event.h"
#include "ppapi/c/pp_point.h"
#include "ppapi/c/pp_rect.h"
#include "ppapi/c/pp_size.h"
#include "ppapi/c/ppb.h"
#include "ppapi/c/ppp.h"
#include "ppapi/c/ppb_audio.h"
#include "ppapi/c/ppb_audio_config.h"
#include "ppapi/c/ppb_core.h"
#include "ppapi/c/ppb_graphics_3d.h"
#include "ppapi/c/ppb_image_data.h"
#include "ppapi/c/ppb_input_event.h"
#include "ppapi/c/ppb_instance.h"
#include "ppapi/c/ppb_opengles2.h"
#include "ppapi/c/ppb_url_loader.h"
#include "ppapi/c/ppp_input_event.h"
#include "ppapi/c/ppp_instance.h"
#include "ppapi/gles2/gl2ext_ppapi.h"

#include <stdint.h>

#include "ppapi/c/pp_bool.h"
#include "ppapi/c/pp_instance.h"
#include "ppapi/c/pp_resource.h"
#include "ppapi/c/ppb.h"

const int kInvalidInstance = 0;
const int kInvalidResource = 0;

struct InstanceInfo
{
	PPB_GetInterface browser_interface;
	PP_Instance instance_id;
	PP_Resource graphics3d_id;
	PP_Resource audio_config_id;
	PP_Resource audio_id;
	int32_t sample_rate;
	int32_t sample_frame_count;
	int32_t frame_counter;
	int32_t width;
	int32_t height;
	int32_t cursor_x;
	int32_t cursor_y;
	volatile int32_t keys_pressed;
};

const PPB_Core* PPBCore=NULL;
const PPB_Graphics3D* PPBGraphics3D=NULL;
const PPB_InputEvent* PPBInputEvent=NULL;
const PPB_Instance* PPBInstance=NULL;
const PPB_MouseInputEvent* PPBMouseInputEvent=NULL;
const PPB_OpenGLES2* PPBOpenGLES2=NULL;

InstanceInfo global;

void RenderFrameStartup(InstanceInfo* instance)
{
	char *ext;
	printf("---- Starting up 3d in NaCl\n");
	ext = (char*) glGetString(GL_EXTENSIONS);
	printf("extensions: %s\n", ext);
	int num_formats;
	glGetIntegerv(GL_NUM_COMPRESSED_TEXTURE_FORMATS, &num_formats);
	printf("number of compressed formats: %d\n", num_formats);
	if (num_formats > 0) {
		int *formats = (int*)alloca(num_formats * sizeof(int));
		const char *fmtstr;
		char temp[1024];
		glGetIntegerv(GL_COMPRESSED_TEXTURE_FORMATS, formats);
		for (int i = 0; i < num_formats; i++) {
			switch(formats[i]) {
				case GL_COMPRESSED_RGB_S3TC_DXT1_EXT:
					fmtstr = "GL_COMPRESSED_RGB_S3TC_DXT1";
					break;
				case GL_COMPRESSED_RGBA_S3TC_DXT1_EXT:
					fmtstr = "GL_COMPRESSED_RGBA_S3TC_DXT1";
					break;
				case GL_COMPRESSED_RGBA_S3TC_DXT3_EXT:
					fmtstr = "GL_COMPRESSED_RGBA_S3TC_DXT3_EXT";
					break;
				case GL_COMPRESSED_RGBA_S3TC_DXT5_EXT:
					fmtstr = "GL_COMPRESSED_RGBA_S3TC_DXT5_EXT";
					break;
				default:
					fmtstr = (const char*) sprintf(temp, "Unknown: 0x%0x", formats[i]);
					break;
			}
			printf("format %d is %s\n", i, fmtstr);
		}
	}
}

void RenderFrame(InstanceInfo* instance)
{
	static int32_t counter = 0;
	glEnable(GL_SCISSOR_TEST);
	glViewport(0, 0, instance->width, instance->height);
	glScissor(0, 0, instance->width, instance->height);
	glClearColor(0.0f, 1.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT);
	glScissor(5, 5, instance->width - 10, instance->height - 10);
	float blue = 1.0f;
	glClearColor(0.0f, 0.0f, blue, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT);
	// Draw a box under the cursor.  Need to flip vertically as gl expects
	// the origin to be the lower right corner.
	const int32_t box_size = 40;
	int32_t box_x = instance->cursor_x - box_size / 2;
	int32_t box_y = instance->height - instance->cursor_y - box_size / 2;
	glScissor(box_x, box_y, box_size, box_size);
	glClearColor(1.0f, instance->keys_pressed ? 1.0f : 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT);
	counter++;
}

void RenderLoop(void* user_data, int32_t result)
{
	InstanceInfo* instance=(InstanceInfo*)user_data;
	glSetCurrentContextPPAPI(instance->graphics3d_id);
	if(instance->frame_counter==0)
	{
		RenderFrameStartup(instance);
	}
	RenderFrame(instance);
	instance->frame_counter++;
	glSetCurrentContextPPAPI(0);
	PP_CompletionCallback cc=PP_MakeCompletionCallback(RenderLoop, instance);
	int32_t swap_result=PPBGraphics3D->SwapBuffers(instance->graphics3d_id, cc);
}

PP_Bool Instance_DidCreate(PP_Instance instance_id, uint32_t argc, const char* argn[], const char* argv[])
{
	assert(global.instance_id==kInvalidInstance);
	global.instance_id=instance_id;
	global.graphics3d_id=kInvalidResource;
	PPBInputEvent->RequestInputEvents(instance_id, PP_INPUTEVENT_CLASS_MOUSE | PP_INPUTEVENT_CLASS_KEYBOARD);
	//global.sample_frame_count=2048;
	//global.sample_rate=PP_AUDIOSAMPLERATE_48000;
	//global.audio._config_id=PPBAudioConfig->CreateStereo16Bit(instance_id, PP_AUDIOSAMPLERATE_48000, global.sample_frame_count);
	//global.audio_id=PPBAudio->Create(instance_id, global.audio_config_id, AudioCallback, &global);
	//PPBAudio->StartPlayback(global.audio_id);

	printf("Module initialized.\n");

	return PP_TRUE;
}

void Instance_DidDestroy(PP_Instance instance_id)
{
	assert(instance_id==global.instance_id);
	if(global.graphics3d_id!=kInvalidResource)
	{
		PPBCore->ReleaseResource(global.graphics3d_id);
		global.graphics3d_id=kInvalidResource;
	}
	global.instance_id=kInvalidInstance;
}

void Instance_DidChangeView(PP_Instance instance_id, const struct PP_Rect* position, const struct PP_Rect* clip)
{
	assert(instance_id==global.instance_id);
	//The first call of an resource, it is set up as invalid.
	if(global.graphics3d_id==kInvalidResource)
	{
		int32_t attribs[]=
		{
			PP_GRAPHICS3DATTRIB_WIDTH, position->size.width,
			PP_GRAPHICS3DATTRIB_HEIGHT, position->size.height,
			PP_GRAPHICS3DATTRIB_DEPTH_SIZE, 24,
			PP_GRAPHICS3DATTRIB_NONE
		};
		global.graphics3d_id=PPBGraphics3D->Create(instance_id, kInvalidResource, attribs);
		int32_t success=PPBInstance->BindGraphics(instance_id, global.graphics3d_id);
		if(success==PP_TRUE)
		{
			PP_CompletionCallback cc=PP_MakeCompletionCallback(RenderLoop, &global);
			PPBCore->CallOnMainThread(0, cc, PP_OK);
		}
		else if (position->size.width != global.width || position->size.height != global.height)
		{
			PPBGraphics3D->ResizeBuffers(global.graphics3d_id, position->size.width, position->size.height);
		}
	}
	global.width=position->size.width;
	global.height=position->size.height;
}

void Instance_DidChangeFocus(PP_Instance instance_id, PP_Bool has_focus)
{
	printf("Instance_DidChangeFocus called.\n");
}

PP_Bool Instance_HandleDocumentLoad(PP_Instance instance_id, PP_Resource pp_url_loader)
{
	printf("Instance_HandleDocumentLoad called.\n");
	return PP_FALSE;
}

PP_Bool InputEvent_HandleEvent(PP_Instance instance_id, PP_Resource input_event)
{
	assert(instance_id==global.instance_id);
	PP_InputEvent_Type type=PPBInputEvent->GetType(input_event);
	switch(type)
	{
		case PP_INPUTEVENT_TYPE_MOUSEMOVE:
			PP_Point point;
			point=PPBMouseInputEvent->GetPosition(input_event);
			global.cursor_x=point.x;
			global.cursor_y=point.y;
			break;
		case PP_INPUTEVENT_TYPE_MOUSEDOWN:
		case PP_INPUTEVENT_TYPE_KEYDOWN:
			global.keys_pressed++;
			break;
		case PP_INPUTEVENT_TYPE_MOUSEUP:
		case PP_INPUTEVENT_TYPE_KEYUP:
			global.keys_pressed--;
			break;
		default:
			break;
	}
	return PP_TRUE;
}



PP_EXPORT int32_t PPP_InitializeModule(PP_Module module, PPB_GetInterface get_browser_interface)
{
	if(!get_browser_interface)
	{
		return -1;
	}
	memset(&global, 0, sizeof(global));
	global.instance_id=kInvalidInstance;
	global.browser_interface=get_browser_interface;
	PPBCore=(const struct PPB_Core*) get_browser_interface(PPB_CORE_INTERFACE);
	PPBInstance=(const struct PPB_Instance*) get_browser_interface(PPB_INSTANCE_INTERFACE);
	PPBOpenGLES2=(const struct PPB_OpenGLES2*) get_browser_interface(PPB_OPENGLES2_INTERFACE);
	PPBInputEvent=(const struct PPB_InputEvent*) get_browser_interface(PPB_INPUT_EVENT_INTERFACE);
	PPBGraphics3D=(const struct PPB_Graphics3D*) get_browser_interface(PPB_GRAPHICS_3D_INTERFACE);
	PPBMouseInputEvent=(const struct PPB_MouseInputEvent*) get_browser_interface(PPB_MOUSE_INPUT_EVENT_INTERFACE);
	if(!PPBCore || !PPBInputEvent || !PPBInstance || !PPBOpenGLES2 || !PPBGraphics3D || !PPBMouseInputEvent)
	{
		return -1;
	}
	if(!glInitializePPAPI(get_browser_interface))
	{
		return -1;
	}
	return PP_OK;
}

PP_EXPORT void PPP_ShutdownModule()
{
	printf("PPP_ShutdownModule called.\n");
	glTerminatePPAPI();
}

PP_EXPORT const void* PPP_GetInterface(const char* interface_name)
{
	static struct PPP_Instance instance_interface=
	{
		&Instance_DidCreate,
		&Instance_DidDestroy,
		&Instance_DidChangeView,
		&Instance_DidChangeFocus,
		&Instance_HandleDocumentLoad
	};
	static struct PPP_InputEvent input_event_interface=
	{
		&InputEvent_HandleEvent
	};
	if(strcmp(interface_name, PPP_INSTANCE_INTERFACE)==0)
	{
		return &instance_interface;
	}
	if(strcmp(interface_name, PPP_INPUT_EVENT_INTERFACE)==0)
	{
		return &input_event_interface;
	}
	return NULL;
}


